import functools
import re
from itertools import starmap
from pathlib import Path
from typing import NamedTuple, Union

from pyhidra import __version__
from pyhidra.constants import GHIDRA_INSTALL_DIR

_APPLICATION_PATTERN = re.compile(r"^application\.(\S+?)=(.*)$")
_APPLICATION_PATH = GHIDRA_INSTALL_DIR / "Ghidra" / "application.properties"


# this is not a NamedTuple as the fields may change
class ApplicationInfo:
    """
    Ghidra Application Properties
    """
    revision_ghidra_src: str = None
    build_date: str = None
    build_date_short: str = None
    name: str
    version: str
    release_name: str
    layout_version: str = None
    gradle_min: str = None
    java_min: str = None
    java_max: str = None
    java_compiler: str = None

    def __init__(self):
        for line in _APPLICATION_PATH.read_text(encoding="utf8").splitlines():
            match = _APPLICATION_PATTERN.match(line)
            if not match:
                continue
            attr = match.group(1).replace('.', '_').replace('-', '_')
            value = match.group(2)
            super().__setattr__(attr, value)

    def __setattr__(self, *attr):
        raise AttributeError(f"cannot assign to field '{attr[0]}'")

    def __delattr__(self, attr):
        raise AttributeError(f"cannot delete field '{attr}'")

    @property
    def extension_path(self) -> Path:
        """
        Path to the user's Ghidra extensions folder
        """
        root = Path.home() / f".{self.name.lower()}"
        return root / f"{root.name}_{self.version}_{self.release_name}" / "Extensions"


_CURRENT_APPLICATION: ApplicationInfo = None
_CURRENT_GHIDRA_VERSION: str = None
MINIMUM_GHIDRA_VERSION = "10.1.1"


def get_current_application() -> ApplicationInfo:
    global _CURRENT_APPLICATION
    if _CURRENT_APPLICATION is None:
        _CURRENT_APPLICATION = ApplicationInfo()
    return _CURRENT_APPLICATION


def get_ghidra_version() -> str:
    global _CURRENT_GHIDRA_VERSION
    if _CURRENT_GHIDRA_VERSION is None:
        _CURRENT_GHIDRA_VERSION = get_current_application().version
    return _CURRENT_GHIDRA_VERSION


_EXTENSION_DEFAULTS: dict = None

def _get_extension_defaults() -> dict:
    global _EXTENSION_DEFAULTS
    if _EXTENSION_DEFAULTS is None:
        _EXTENSION_DEFAULTS = {
            "name":  "pyhidra",
            "description":  "Native Python Plugin",
            "author":  "Department of Defense Cyber Crime Center (DC3)",
            "createdOn":  "",
            "version": get_ghidra_version(),
            "pyhidra": __version__
        }
    return _EXTENSION_DEFAULTS

def _properties_wrapper(cls):
    @functools.wraps(cls)
    def wrapper(ext: Union[Path, dict] = None):
        if isinstance(ext, dict):
            return cls(**ext)
        def cast(key, value):
            # __annotations__ is created for NamedTuple since its first implementation
            return cls.__annotations__[key](value)

        if ext is None:
            return cls(_get_extension_defaults())
        lines = ext.read_text().splitlines()
        args = tuple(starmap(cast, map(lambda l: l.split('='), lines)))
        return cls(*args)
    return wrapper


@_properties_wrapper
class ExtensionDetails(NamedTuple):
    """
    Python side ExtensionDetails
    """

    name: str = "pyhidra"
    description: str = "Native Python Plugin"
    author: str = "Department of Defense Cyber Crime Center (DC3)"
    createdOn: str = ""
    version: str = None
    pyhidra: str = __version__

    def __repr__(self):
        cls = self.__class__
        return '\n'.join(starmap(lambda i, k: f"{k}={self[i]}", enumerate(cls.__annotations__)))



from pathlib import Path

from java.lang import ClassLoader
from utility.application import ApplicationLayout

from pyhidra.java.plugin.plugin import PyPhidraPlugin
from pyhidra.javac import java_compile
from pyhidra.version import get_current_application, ExtensionDetails



PACKAGE = "dc3.pyhidra.plugin"
PLUGIN_NAME = "pyhidra"


def _get_extension_details(layout: ApplicationLayout):
    return ExtensionDetails(
        PLUGIN_NAME,
        "Native Python Plugin",
        "Department of Defense Cyber Crime Center (DC3)",
        "",
        layout.getApplicationProperties().getApplicationVersion()
    )


def install():
    """
    Install the plugin in Ghidra
    """
    path = get_current_application().extension_path / "pyhidra"
    ext = path / "extension.properties"
    manifest = path / "Module.manifest"
    if not manifest.exists():
        jar_path = path / "lib" / (PLUGIN_NAME + ".jar")
        java_compile(Path(__file__).parent.parent, jar_path)
        ClassLoader.getSystemClassLoader().addPath(jar_path.absolute())

    if not manifest.exists():
        ext.write_text(str(ExtensionDetails()))
        # required empty file, might be usable for version control in the future
        manifest.touch()

    PyPhidraPlugin.register()



import contextlib
import itertools
import rlcompleter
import sys
import threading
from code import InteractiveConsole

from ghidra.app.plugin.core.console import CodeCompletion
from ghidra.app.plugin.core.interpreter import InterpreterConnection, InterpreterPanelService
from ghidra.util.task import DummyCancellableTaskMonitor
from java.io import BufferedReader, InputStreamReader, PushbackReader
from java.lang import ClassLoader, Runnable, String
from java.util.function import Consumer
from jpype import JClass, JImplements, JOverride
from resources import ResourceManager
from utility.function import Callback

from pyhidra.java.plugin.completions import PythonCodeCompleter
from pyhidra.script import PyGhidraScript


def _set_field(cls, fname, value, obj=None):
    cls = cls.class_
    field = cls.getDeclaredField(fname)
    field.setAccessible(True)
    field.set(obj, value)
    field.setAccessible(False)


def _run_script(script):
    PyGhidraScript(script).run()


class PyConsole(InteractiveConsole):
    """
    Pyhidra Interactive Console
    """

    def __init__(self, py_plugin) -> None:
        super().__init__(locals=PyGhidraScript(py_plugin.script))
        self._plugin = py_plugin
        console = py_plugin.service.createInterpreterPanel(py_plugin, False)
        self._console = console
        self._reader = PushbackReader(InputStreamReader(console.getStdin()))
        self._line_reader = BufferedReader(self._reader)
        self._out = console.getOutWriter()
        self._err = console.getErrWriter()
        self._writer = self._out
        self._thread = None
        state = self.locals._script.getState()
        self.locals._script.set(state, DummyCancellableTaskMonitor(), console.getOutWriter())
        console.addFirstActivationCallback(Callback @ self.interact)

    def interact(self, banner=None, exitmsg=None):
        from ghidra.framework import Application
        version = Application.getApplicationVersion()
        name = Application.getApplicationReleaseName()
        banner = f"Python Interpreter for Ghidra {version} {name}\n"
        banner += f"Python {sys.version} on {sys.platform}"
        target = super().interact
        targs = {'banner': banner}
        self._thread = threading.Thread(target=target, name="Interpreter", kwargs=targs)
        self._thread.start()

    def raw_input(self, prompt=''):
        self._console.setPrompt(prompt)
        c = self._reader.read()
        if c == -1:
            raise EOFError
        if c == ord('\n'):
            return '\n'
        self._reader.unread(c)
        return self._line_reader.readLine()

    def write(self, data: str):
        self._writer.write(String @ data)
        self._writer.flush()

    def dispose(self):
        """
        Release the console resources
        """
        self._console.dispose()
        if self._thread is not None and self._thread.is_alive():
            self._thread.join()

    def showsyntaxerror(self, filename=None):
        self._writer = self._err
        super().showsyntaxerror(filename=filename)
        self._writer = self._out

    def showtraceback(self) -> None:
        self._writer = self._err
        super().showtraceback()
        self._writer = self._out

    @contextlib.contextmanager
    def _run_context(self):
        transaction = -1
        success = False
        program = self._plugin.program
        if program is not None:
            transaction = program.startTransaction("Python command")
        try:
            with contextlib.redirect_stdout(self._out), contextlib.redirect_stderr(self._err):
                yield
                success = True
        finally:
            if transaction != -1:
                program = self._plugin.program
                if program is not None:
                    program.endTransaction(transaction, success)

    def runcode(self, code):
        with self._run_context():
            super().runcode(code)
        self._out.flush()
        self._err.flush()


@JImplements(InterpreterConnection)
class PyPhidraPlugin:
    """
    The Python side PyhidraPlugin
    """

    # pylint: disable=missing-function-docstring, invalid-name

    def __init__(self, plugin):
        if hasattr(self, '_plugin'):
            # this gets entered twice for some reason
            return
        self._plugin = plugin
        gcl = ClassLoader.getSystemClassLoader()
        plugin_cls = JClass("dc3.pyhidra.plugin.PyhidraPlugin", loader=gcl)
        _set_field(plugin_cls, "finalizer", Runnable @ self.dispose, plugin)
        self.console = PyConsole(self)
        self.completer = PythonCodeCompleter(self.console)

    @staticmethod
    def register():
        gcl = ClassLoader.getSystemClassLoader()
        plugin = JClass("dc3.pyhidra.plugin.PyhidraPlugin", loader=gcl)
        provider = JClass("dc3.pyhidra.plugin.PyScriptProvider", loader=gcl)
        _set_field(plugin, "initializer", Consumer @ PyPhidraPlugin)
        _set_field(provider, "scriptRunner", Consumer @ _run_script)

    def _set_plugin(self, plugin):
        self._plugin = plugin

    def dispose(self):
        """
        Release the plugin resources
        """
        self.console.dispose()

    def _gen_completions(self, cmd: str):
        completer = rlcompleter.Completer(namespace=self)
        for state in itertools.count():
            completion = completer.complete(cmd, state)
            if completion is None:
                break
            yield CodeCompletion(cmd, completion, None)

    @property
    def program(self):
        return self._plugin.getCurrentProgram()

    @property
    def script(self):
        return self._plugin.script

    @property
    def service(self):
        return self._plugin.getTool().getService(InterpreterPanelService.class_)

    @JOverride
    def getCompletions(self, cmd: str):
        return self.completer.get_completions(cmd)

    @JOverride
    def getIcon(self):
        return ResourceManager.loadImage("images/python.png")

    @JOverride
    def getTitle(self):
        return "Pyhidra"



import os
import pathlib

GHIDRA_INSTALL_DIR = pathlib.Path(os.environ["GHIDRA_INSTALL_DIR"])
LAUNCH_PROPERTIES = GHIDRA_INSTALL_DIR / "support" / "launch.properties"
UTILITY_JAR = GHIDRA_INSTALL_DIR / "Ghidra" / "Framework" / "Utility" / "lib" / "Utility.jar"
LAUNCHSUPPORT = GHIDRA_INSTALL_DIR / "support" / "LaunchSupport.jar"



import importlib
import importlib.machinery
import importlib.util
import inspect
import logging
import sys
import traceback
from collections.abc import ItemsView, KeysView
from jpype import JClass, JImplementationFor


_NO_ATTRIBUTE = object()


class _StaticMap(dict):

    __slots__ = ('script',)

    def __init__(self, script: "PyGhidraScript"):
        super().__init__()
        self.script = script

    def __getitem__(self, key):
        res = self.get(key, _NO_ATTRIBUTE)
        if res is not _NO_ATTRIBUTE:
            return res
        raise KeyError(key)

    def get(self, key, default=None):
        res = self.script.get_static(key)
        return res if res is not _NO_ATTRIBUTE else default

    def __iter__(self):
        yield from self.script

    def keys(self):
        return KeysView(self)

    def items(self):
        return ItemsView(self)


class _JavaProperty(property):

    def __init__(self, field):
        super().__init__()
        self._field = field

    def __get__(self, obj, cls):
        return self._field.fget(obj)

    def __set__(self, obj, value):
        self._field.fset(obj, value)


#pylint: disable=too-few-public-methods
@JImplementationFor("dc3.pyhidra.plugin.PythonFieldExposer")
class _PythonFieldExposer:

    #pylint: disable=no-member
    def __jclass_init__(self):
        exposer = JClass("dc3.pyhidra.plugin.PythonFieldExposer")
        if self.class_ == exposer:
            return
        try:
            for k, v in exposer.getProperties(self.class_).items():
                self._customize(k, _JavaProperty(v))
        # allowing any exception to escape here may cause the jvm to terminate
        # pylint: disable=bare-except
        except:
            logger = logging.getLogger(__name__)
            logger.error("Failed to add property customizations for %s", self, exc_info=1)


class _GhidraScriptModule:

    def __init__(self, spec: importlib.machinery.ModuleSpec):
        super().__setattr__("__dict__", spec.loader_state["script"])

    def __setattr__(self, attr, value):
        if hasattr(self, attr):
            raise AttributeError(f"readonly attribute {attr}")
        super().__setattr__(attr, value)


class _GhidraScriptLoader(importlib.machinery.SourceFileLoader):

    def __init__(self, script: "PyGhidraScript", spec: importlib.machinery.ModuleSpec):
        super().__init__(spec.name, spec.origin)
        spec.loader_state = {"script": script}

    def create_module(self, spec: importlib.machinery.ModuleSpec):
        return _GhidraScriptModule(spec)


# pylint: disable=missing-function-docstring
class PyGhidraScript(dict):
    """
    Python GhidraScript Wrapper
    """

    def __init__(self, jobj=None):
        super().__init__()
        if jobj is None:
            jobj = JClass("dc3.pyhidra.plugin.PyScriptProvider").PyhidraHeadlessScript()
        self._script = jobj

        # ensure the builtin set takes priority over GhidraScript.set
        super().__setitem__("set", set)

        # ensure that GhidraScript.print is used for print
        # so the output goes to the expected console
        super().__setitem__("print", self._print_wrapper())

        super().__setitem__("__this__", self._script)

    def _print_wrapper(self):
        def _print(*objects, sep=' ', end='\n', file=None, flush=False):
            if file is None:
                file = self._script.writer
            print(*objects, sep=sep, end=end, file=file, flush=flush)
        _print.__doc__ = print.__doc__
        return _print

    def __missing__(self, k):
        attr = getattr(self._script, k, _NO_ATTRIBUTE)
        if attr is not _NO_ATTRIBUTE:
            return attr
        raise KeyError(k)

    def __setitem__(self, k, v):
        attr = inspect.getattr_static(self._script, k, _NO_ATTRIBUTE)
        if attr is not _NO_ATTRIBUTE and isinstance(attr, property):
            setattr(self._script, k, v)
        else:
            super().__setitem__(k, v)

    def __iter__(self):
        yield from super().__iter__()
        yield from dir(self._script)

    def get_static(self, key):
        res = self.get(key, _NO_ATTRIBUTE)
        if res is not _NO_ATTRIBUTE:
            return res
        return inspect.getattr_static(self._script, key, _NO_ATTRIBUTE)

    def get_static_view(self):
        return _StaticMap(self)

    def set(self, state, monitor, writer):
        """
        see GhidraScript.set
        """
        self._script.set(state, monitor, writer)

    def run(self, script_path: str = None, script_args: list = None):
        """
        Run this GhidraScript

        :param script_path: The path of the python script
        :param script_args: The arguments for the python script
        """
        sf = self._script.getSourceFile()
        if sf is None and script_path is None:
            return
        if script_path is None:
            script_path = sf.getAbsolutePath()
            script_args = self._script.getScriptArgs()

        if script_args is None:
            script_args = []

        orig_argv = sys.argv
        try:
            # Temporarily set command line arguments.
            sys.argv = [script_path] + list(script_args)

            spec = importlib.util.spec_from_file_location('__main__', script_path)
            spec.loader = _GhidraScriptLoader(self, spec)
            m = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(m)
            # pylint: disable=bare-except
            except:
                # filter the traceback so that it stops at the script
                exc_type, exc_value, exc_tb = sys.exc_info()
                i = 0
                tb = traceback.extract_tb(exc_tb)
                for fs in tb:
                    if fs.filename == script_path:
                        break
                    i += 1
                ss = traceback.StackSummary.from_list(tb[i:])
                e = traceback.TracebackException(exc_type, exc_value, exc_tb)
                e.stack = ss
                self._script.printerr(''.join(e.format()))
        finally:
            sys.argv = orig_argv



import contextlib
import logging
import platform
import re
import shutil
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import NoReturn

import jpype
from jpype import imports

from . import __version__
from .constants import LAUNCH_PROPERTIES, LAUNCHSUPPORT, GHIDRA_INSTALL_DIR, UTILITY_JAR
from .version import get_current_application, get_ghidra_version, MINIMUM_GHIDRA_VERSION, \
    ExtensionDetails


_GET_JAVA_HOME = f'java -cp "{LAUNCHSUPPORT}" LaunchSupport "{GHIDRA_INSTALL_DIR}" -jdk_home -save'


def _jvm_args():
    suffix = "_" + platform.system().upper()
    option_pattern: re.Pattern = re.compile(fr"VMARGS(?:{suffix})?=(.+)")
    properties = []
    with open(LAUNCH_PROPERTIES, "r", encoding='utf-8') as fd:
        # this file is small so just read it at once
        for line in fd.readlines():
            match = option_pattern.match(line)
            if match:
                properties.append(match.group(1))

    # even though ignoreUnrecognized is True when starting the VM this is still needed
    properties.insert(0, "-XX:+IgnoreUnrecognizedVMOptions")

    return properties


@contextlib.contextmanager
def _silence_java_output(stdout=True, stderr=True):
    from java.io import OutputStream, PrintStream
    from java.lang import System
    out = System.out
    err = System.err
    null = PrintStream(OutputStream.nullOutputStream())

    # The user's Java SecurityManager might not allow this
    with contextlib.suppress(jpype.JException):
        if stdout:
            System.setOut(null)
        if stderr:
            System.setErr(null)

    try:
        yield
    finally:
        with contextlib.suppress(jpype.JException):
            System.setOut(out)
            System.setErr(err)

def _get_libjvm_path(java_home: Path) -> Path:
    for p in java_home.glob("*/server/*jvm.*"):
        if p.suffix != ".debuginfo":
            return p


class PyhidraLauncher:
    """
    Base pyhidra launcher
    """

    def __init__(self, verbose):
        self.verbose = verbose
        self.java_home = None
        self.class_path = [str(UTILITY_JAR)]
        self.vm_args = _jvm_args()
        self.layout = None
        self.args = []

    def add_classpaths(self, *args):
        """
        Add additional entries to the classpath when starting the JVM
        """
        self.class_path += args

    def add_vmargs(self, *args):
        """
        Add additional vmargs for launching the JVM
        """
        self.vm_args += args

    @classmethod
    def _report_fatal_error(cls, title: str, msg: str) -> NoReturn:
        sys.exit(f"{title}: {msg}")

    @classmethod
    def _update(cls):
        ext = get_current_application().extension_path / "pyhidra" / "extension.properties"
        if ext.exists():
            details = ExtensionDetails(ext)
            if details.pyhidra < __version__:
                # delete the existing extension so it will be up-to-date
                try:
                    shutil.rmtree(ext.parent)
                except:  # pylint: disable=bare-except
                    title = "Plugin Update Failed"
                    msg = f"Could not delete existing plugin at\n{ext.parent}"
                    logging.exception(msg)
                    cls._report_fatal_error(title, msg)

    @classmethod
    def check_ghidra_version(cls):
        """
        Checks if the currently installed Ghidra version is supported.
        The launcher will report the problem and terminate if it is not supported.
        """
        if get_ghidra_version() < MINIMUM_GHIDRA_VERSION:
            cls._report_fatal_error(
                "Unsupported Version",
                textwrap.dedent(f"""\
                    Ghidra version {get_ghidra_version()} is not supported
                    The minimum required version is {MINIMUM_GHIDRA_VERSION}
                """).rstrip()
            )

    def start(self):
        """
        Starts Jpype connection to Ghidra (if not already started).
        """
        if not jpype.isJVMStarted():

            self.check_ghidra_version()

            if self.java_home is None:
                java_home = subprocess.check_output(_GET_JAVA_HOME, encoding="utf-8", shell=True)
                self.java_home = Path(java_home.rstrip())

            jvm = _get_libjvm_path(self.java_home)

            jpype.startJVM(
                str(jvm),
                *self.vm_args,
                ignoreUnrecognized=True,
                convertStrings=True,
                classpath=self.class_path
            )
            imports.registerDomain("ghidra")

            from ghidra import GhidraLauncher

            self._update()

            self.layout = GhidraLauncher.initializeGhidraEnvironment()

            from pyhidra.java.plugin import install

            install()

            # import properties to register the property customizer
            from . import properties as _

            self._launch()

    def _launch(self):
        pass

    @staticmethod
    def has_launched() -> bool:
        """
        Checks if jpype has started and if Ghidra has been fully initialized.
        """
        if not jpype.isJVMStarted():
            return False

        from ghidra.framework import Application
        return Application.isInitialized()


class DeferredPyhidraLauncher(PyhidraLauncher):
    """
    PyhidraLauncher which allows full Ghidra initialization to be deferred.
    initialize_ghidra must be called before all Ghidra classes are fully available.
    """

    def __init__(self, verbose=False):
        super().__init__(verbose)

    def initialize_ghidra(self, headless=True):
        """
        Finished Ghidra initialization

        :param headless: whether or not to initialize Ghidra in headless mode.
            (Defaults to True)
        """
        from ghidra import GhidraRun
        from ghidra.framework import Application, HeadlessGhidraApplicationConfiguration
        with _silence_java_output(not self.verbose, not self.verbose):
            if headless:
                config = HeadlessGhidraApplicationConfiguration()
                Application.initializeApplication(self.layout, config)
            else:
                GhidraRun().launch(self.layout, self.args)


class HeadlessPyhidraLauncher(PyhidraLauncher):
    """
    Headless pyhidra launcher
    """

    def __init__(self, verbose=False):
        super().__init__(verbose)

    def _launch(self):
        from ghidra.framework import Application, HeadlessGhidraApplicationConfiguration
        with _silence_java_output(not self.verbose, not self.verbose):
            config = HeadlessGhidraApplicationConfiguration()
            Application.initializeApplication(self.layout, config)


def _popup_error(header: str, msg: str) -> NoReturn:
    import tkinter.messagebox
    tkinter.messagebox.showerror(header, msg)
    sys.exit(msg)


class GuiPyhidraLauncher(PyhidraLauncher):
    """
    GUI pyhidra launcher
    """

    def __init__(self, verbose=False):
        super().__init__(verbose)

    @classmethod
    def _report_fatal_error(cls, title: str, msg: str) -> NoReturn:
        _popup_error(title, msg)

    @staticmethod
    def _get_thread(name: str):
        from java.lang import Thread
        for t in Thread.getAllStackTraces().keySet():
            if t.getName() == name:
                return t
        return None

    def _launch(self):
        import ctypes
        from ghidra import GhidraRun

        if sys.platform == "win32":
            appid = ctypes.c_wchar_p(get_current_application().name)
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)
        jpype.setupGuiEnvironment(lambda: GhidraRun().launch(self.layout, self.args))
        try:
            t = GuiPyhidraLauncher._get_thread("main")
            if t is not None:
                try:
                    t.join()
                except RuntimeError as e:
                    if "java.lang.InterruptedException" not in e.args:
                        raise
        finally:
            jpype.shutdownGuiEnvironment()



import contextlib
import pathlib
from typing import Union, TYPE_CHECKING, Tuple, ContextManager, List

from pyhidra.converters import *  # pylint: disable=wildcard-import, unused-wildcard-import


if TYPE_CHECKING:
    import ghidra


def start(verbose=False):
    """
    Starts the JVM and loads the Ghidra libraries.
    Full Ghidra initialization is deferred.

    :param verbose: Enable verbose output during JVM startup (Defaults to False)
    :return: The DeferredPhyidraLauncher used to start the JVM
    """
    from pyhidra.launcher import HeadlessPyhidraLauncher
    launcher = HeadlessPyhidraLauncher(verbose=verbose)
    launcher.start()
    return launcher


def _setup_project(
        binary_path: Union[str, pathlib.Path],
        project_location: Union[str, pathlib.Path] = None,
        project_name: str = None
) -> Tuple["ghidra.base.project.GhidraProject", "ghidra.program.model.listing.Program"]:
    from ghidra.base.project import GhidraProject
    from java.io import IOException
    if binary_path is not None:
        binary_path = pathlib.Path(binary_path)
    if project_location:
        project_location = pathlib.Path(project_location)
    else:
        project_location = binary_path.parent
    if not project_name:
        project_name = f"{binary_path.name}_ghidra"
    project_location = project_location / project_name
    project_location.mkdir(exist_ok=True, parents=True)

    # Open/Create project
    program = None
    try:
        project = GhidraProject.openProject(project_location, project_name, True)
        if binary_path is not None:
            if project.getRootFolder().getFile(binary_path.name):
                program = project.openProgram("/", binary_path.name, False)
    except IOException:
        project = GhidraProject.createProject(project_location, project_name, False)

    if binary_path is not None and program is None:
        program = project.importProgram(binary_path)
        project.saveAs(program, "/", program.getName(), True)

    return project, program


def _setup_script(project, program):
    from pyhidra.script import PyGhidraScript
    from ghidra.app.script import GhidraState
    from ghidra.program.util import ProgramLocation
    from ghidra.util.task import TaskMonitor

    from java.io import PrintWriter
    from java.lang import System

    if project is not None:
        project = project.getProject()

    location = None
    if program is not None:
        # create a GhidraState and setup a HeadlessScript with it
        mem = program.getMemory().getLoadedAndInitializedAddressSet()
        if not mem.isEmpty():
            location = ProgramLocation(program, mem.getMinAddress())
    state = GhidraState(None, project, program, location, None, None)
    script = PyGhidraScript()
    script.set(state, TaskMonitor.DUMMY, PrintWriter(System.out))
    return script


@contextlib.contextmanager
def open_program(
        binary_path: Union[str, pathlib.Path],
        project_location: Union[str, Path] = None,
        project_name: str = None,
        analyze=True
) -> ContextManager["ghidra.program.flatapi.FlatProgramAPI"]:
    """
    Opens given binary path in Ghidra and returns FlatProgramAPI object.

    :param binary_path:
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra")
    :param analyze: Whether to run analysis before returning.
    :return: A Ghidra FlatProgramAPI object.
    """

    from pyhidra.launcher import PyhidraLauncher, HeadlessPyhidraLauncher

    if not PyhidraLauncher.has_launched():
        HeadlessPyhidraLauncher().start()

    from ghidra.program.flatapi import FlatProgramAPI

    project, program = _setup_project(binary_path, project_location, project_name)

    try:
        flat_api = FlatProgramAPI(program)

        if analyze:
            from ghidra.program.util import GhidraProgramUtilities
            from ghidra.app.script import GhidraScriptUtil
            if GhidraProgramUtilities.shouldAskToAnalyze(program):
                GhidraScriptUtil.acquireBundleHostReference()
                try:
                    flat_api.analyzeAll(program)
                finally:
                    GhidraScriptUtil.releaseBundleHostReference()
        yield flat_api
    finally:
        project.save(program)
        project.close()


@contextlib.contextmanager
def _flat_api(
        binary_path: Union[str, Path],
        project_location: Union[str, Path] = None,
        project_name: str = None,
        verbose=False
):
    """
    Runs a given script on a given binary path.

    :param binary_path: Path to binary file.
    :param script_path: Path to script to run.
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra")
    :param script_args: Command line arguments to pass to script.
    :param verbose: Enable verbose output during Ghidra initialization.
    """
    from pyhidra.launcher import PyhidraLauncher, HeadlessPyhidraLauncher

    if not PyhidraLauncher.has_launched():
        HeadlessPyhidraLauncher(verbose=verbose).start()

    project, program = None, None
    if binary_path or project_location:
        project, program = _setup_project(binary_path, project_location, project_name)

    try:
        yield _setup_script(project, program)
    finally:
        if project is not None:
            if program is not None:
                project.save(program)
            project.close()


# pylint: disable=too-many-arguments
def run_script(
    binary_path: Union[str, Path],
    script_path: Union[str, Path],
    project_location: Union[str, Path] = None,
    project_name: str = None,
    script_args: List[str] = None,
    verbose=False
):
    """
    Runs a given script on a given binary path.

    :param binary_path: Path to binary file, may be None
    :param script_path: Path to script to run.
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file if not None)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra" if not None)
    :param script_args: Command line arguments to pass to script.
    :param verbose: Enable verbose output during Ghidra initialization.
    """
    script_path = str(script_path)
    with _flat_api(binary_path, project_location, project_name, verbose) as script:
        script.run(script_path, script_args)




__version__ = "0.1.4"

# Expose API
from .ghidra import run_script, start, open_program
from .gui import get_current_interpreter
from .launcher import DeferredPyhidraLauncher, HeadlessPyhidraLauncher, GuiPyhidraLauncher


__all__ = [
    "run_script", "start", "open_program",
    "DeferredPyhidraLauncher", "HeadlessPyhidraLauncher", "GuiPyhidraLauncher"
]



import os
import sys


def gui():
    """
    Starts the Ghidra GUI
    """
    if not "GHIDRA_INSTALL_DIR" in os.environ:
        import tkinter.messagebox
        msg = "GHIDRA_INSTALL_DIR not set.\nPlease see the README for setup instructions"
        tkinter.messagebox.showerror("Improper Setup", msg)
        sys.exit()
    import pyhidra
    pyhidra.GuiPyhidraLauncher().start()


def get_current_interpreter():
    """
    Gets the underlying GhidraScript for the focused Pyhidra InteractiveConsole.
    This will always return None unless it is being access from a function
    called from within the interactive console.

    :return: The GhidraScript for the active interactive console.
    """
    
    try:
        from ghidra.framework.main import AppInfo
        project = AppInfo.getActiveProject()
        if project is None:
            return None
        ts = project.getToolServices()
        tool = None
        for t in ts.getRunningTools():
            if t.getActiveWindow().isFocused():
                tool = t
                break
        if tool is None:
            return None
        for plugin in tool.getManagedPlugins():
            if plugin.name == 'PyhidraPlugin':
                return plugin.script
    except ImportError:
        return None



import argparse
import code
import sys
from pathlib import Path

import pyhidra
import pyhidra.gui


def _create_shortcut():
    from pyhidra.win_shortcut import create_shortcut
    create_shortcut(Path(sys.argv[-1]))


def _interpreter(interpreter_globals: dict):
    from ghidra.framework import Application
    version = Application.getApplicationVersion()
    name = Application.getApplicationReleaseName()
    banner = f"Python Interpreter for Ghidra {version} {name}\n"
    banner += f"Python {sys.version} on {sys.platform}"
    code.interact(banner=banner, local=interpreter_globals, exitmsg='')


# pylint: disable=too-few-public-methods
class PyhidraArgs(argparse.Namespace):
    """
    Custom namespace for holding the command line arguments
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.verbose = False
        self.binary_path = None
        self.script_path = None
        self.project_name = None
        self.project_path = None
        self.script_args = []

    def func(self):
        """
        Run script or enter repl
        """
        if self.script_path is not None:
            pyhidra.run_script(
                self.binary_path,
                self.script_path,
                project_location=self.project_path,
                project_name=self.project_name,
                script_args=self.script_args,
                verbose=self.verbose
            )
        elif self.binary_path is not None:
            from .ghidra import _flat_api
            args = self.binary_path, self.project_path, self.project_name, self.verbose
            with _flat_api(*args) as api:
                _interpreter(api)
        else:
            pyhidra.HeadlessPyhidraLauncher(verbose=self.verbose).start()
            _interpreter(globals())


class PathAction(argparse.Action):
    """
    Custom action for handling script and binary paths as positional arguments
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nargs = '*'
        self.help = "Headless script and/or binary path. "\
            "If neither are provided pyhidra will drop into a repl."
        self.type = Path

    def __call__(self, parser, namespace, values, option_string=None):
        count = 0
        for p in values:
            if p.exists() and p.is_file():
                if p.suffix == ".py":
                    if namespace.script_path is not None:
                        # assume an additional script is meant to be a parameter to the first one
                        break
                    namespace.script_path = p
                else:
                    if namespace.binary_path is not None:
                        if namespace.script_path is None:
                            raise ValueError("binary_path specified multiple times")
                        # assume it is a script parameter
                        break
                    namespace.binary_path = p
                count += 1
            else:
                break
            if count > 1:
                break
        values[:] = values[count:]

def _get_parser():
    parser = argparse.ArgumentParser(prog="pyhidra")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Enable verbose output during Ghidra initialization"
    )
    parser.add_argument(
        "-g",
        "--gui",
        action="store_const",
        dest="func",
        const=pyhidra.gui.gui,
        help="Start Ghidra GUI"
    )
    if sys.platform == "win32":
        parser.add_argument(
            "-s",
            "--shortcut",
            action="store_const",
            dest="func",
            const=_create_shortcut,
            help="Creates a shortcut that can be pinned to the taskbar (Windows only)"
        )
    parser.add_argument(
        "script_path | binary_path",
        metavar="script | binary",
        action=PathAction
    )
    parser.add_argument(
        "script_args",
        help="Arguments to be passed to the headless script",
        nargs=argparse.REMAINDER
    )
    parser.add_argument(
        "--project-name",
        type=str,
        dest="project_name",
        metavar="name",
        help="Project name to use. "
        "(defaults to binary filename with \"_ghidra\" suffix if provided else None)"
    )
    parser.add_argument(
        "--project-path",
        type=Path,
        dest="project_path",
        metavar="path",
        help="Location to store project. "
        "(defaults to same directory as binary file if provided else None)"
    )
    return parser


def main():
    """
    pyhidra module main function
    """
    _get_parser().parse_args(namespace=PyhidraArgs()).func()


if __name__ == "__main__":
    main()
